# index.py
''' Content indexer '''

from __future__ import absolute_import, with_statement

import os
import logging

import watchdog.observers
import watchdog.events

from . import entry
from . import model

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

ENTRY_TYPES = ['.md', '.htm', '.html']


def scan_file(fullpath, relpath, assign_id):
    """ Scan a file for the index

    fullpath -- The full path to the file
    relpath -- The path to the file, relative to its base directory
    assign_id -- Whether to assign an ID to the file if not yet assigned
    """

    _, ext = os.path.splitext(fullpath)

    try:
        if ext in ENTRY_TYPES:
            return entry.scan_file(fullpath, relpath, assign_id)

        return True
    except:  # pylint: disable=bare-except
        logger.exception("Got error parsing %s", fullpath)
    return None


def get_last_mtime(fullpath):
    """ Get the last known modification time for a file """
    record = model.FileMTime.get_or_none(model.FileMTime.file_path == fullpath)
    if record:
        return record.stat_mtime
    return None


def set_last_mtime(fullpath, mtime=None):
    """ Set the last known modification time for a file """
    try:
        mtime = mtime or os.stat(fullpath).st_mtime

        record, created = model.FileMTime.get_or_create(
            file_path=fullpath, defaults={'stat_mtime': os.stat(fullpath).st_mtime})
        if not created:
            record.stat_mtime = mtime
            record.save()
    except FileNotFoundError:
        model.FileMTime.delete().where(model.FileMTime.file_path == fullpath)


class IndexWatchdog(watchdog.events.FileSystemEventHandler):
    """ Watchdog handler """

    def __init__(self, content_dir):
        self.content_dir = content_dir

    def update_file(self, fullpath):
        """ Update a file """
        relpath = os.path.relpath(fullpath, self.content_dir)

        try:
            if scan_file(fullpath, relpath, True):
                logger.info("Updated %s", fullpath)
                set_last_mtime(fullpath)
            else:
                logger.warning("Couldn't update %s", fullpath)
        except:  # pylint: disable=bare-except
            logger.exception("Got error updating %s", fullpath)

    def on_created(self, event):
        """ on_created handler """
        logger.info("file created: %s", event.src_path)
        if not event.is_directory:
            self.update_file(event.src_path)

    def on_modified(self, event):
        """ on_modified handler """
        logger.info("file modified: %s", event.src_path)
        if not event.is_directory:
            self.update_file(event.src_path)

    def on_moved(self, event):
        """ on_moved handler """
        logger.info("file moved: %s -> %s", event.src_path, event.dest_path)
        if not event.is_directory:
            self.update_file(event.src_path)
            self.update_file(event.dest_path)

    def on_deleted(self, event):
        """ on_deleted handler """
        logger.info("File deleted: %s", event.src_path)
        if not event.is_directory:
            self.update_file(event.src_path)


def background_scan(content_dir):
    """ Start background scanning a directory for changes """
    observer = watchdog.observers.Observer()
    observer.schedule(IndexWatchdog(content_dir),
                      content_dir, recursive=True)
    logging.info("Watching %s for changes", content_dir)
    observer.start()


def scan_index(content_dir):
    """ Scan all files in a content directory """
    fixups = []
    for root, _, files in os.walk(content_dir, followlinks=True):
        for file in files:
            fullpath = os.path.join(root, file)
            relpath = os.path.relpath(fullpath, content_dir)

            mtime = os.stat(fullpath).st_mtime
            last_mtime = get_last_mtime(fullpath)
            if not last_mtime or last_mtime < mtime:
                if scan_file(fullpath, relpath, False):
                    logger.info("Scanned %s", fullpath)
                    set_last_mtime(fullpath, mtime)
                else:
                    # file scan failed, add to the fixups queue
                    fixups.append((fullpath, relpath))
                    logger.info("Scheduling fixup for %s", fullpath)

    # perform the fixup queue
    for fullpath, relpath in fixups:
        if scan_file(fullpath, relpath, True):
            logger.info("Fixed up %s", fullpath)
        else:
            logger.warning("Couldn't fix up %s", fullpath)



# queries.py
""" Collection of commonly-used queries """

from __future__ import absolute_import, with_statement

import arrow

from . import model, utils


def where_entry_visible(date=None):
    """ Generate a where clause for currently-visible entries

    Arguments:

    date -- The date to generate it relative to (defaults to right now)
    """

    return (
        (model.Entry.status == model.PublishStatus.PUBLISHED) |
        (
            (model.Entry.status == model.PublishStatus.SCHEDULED) &
            (model.Entry.entry_date <= (date or arrow.utcnow().datetime))
        )
    )


def where_entry_visible_future():
    """ Generate a where clause for entries that are visible now or in the future """
    return (
        (model.Entry.status == model.PublishStatus.PUBLISHED) |
        (model.Entry.status == model.PublishStatus.SCHEDULED)
    )


def where_entry_category(category, recurse=False):
    """ Generate a where clause for a particular category """

    if category or not recurse:
        cat_where = (model.Entry.category == str(category))

        if recurse:
            # We're recursing and aren't in /, so add the prefix clause
            cat_where = cat_where | (
                model.Entry.category.startswith(str(category) + '/'))
    else:
        cat_where = True

    return cat_where


def where_before_entry(entry):
    """ Generate a where clause for prior entries

    entry -- The entry of reference
    """
    return (model.Entry.entry_date < entry.entry_date) | (
        (model.Entry.entry_date == entry.entry_date) &
        (model.Entry.id < entry.id)
    )


def where_after_entry(entry):
    """ Generate a where clause for later entries

    entry -- the entry of reference
    """
    return (model.Entry.entry_date > entry.entry_date) | (
        (model.Entry.entry_date == entry.entry_date) &
        (model.Entry.id > entry.id)
    )


def where_entry_last(entry):
    """ Generate a where clause where this is the last entry

    entry -- the entry of reference
    """
    return (model.Entry.entry_date < entry.entry_date) | (
        (model.Entry.entry_date == entry.entry_date) &
        (model.Entry.id <= entry.id)
    )


def where_entry_first(entry):
    """ Generate a where clause where this is the first entry

    entry -- the entry of reference
    """
    return (model.Entry.entry_date > entry.entry_date) | (
        (model.Entry.entry_date == entry.entry_date) &
        (model.Entry.id >= entry.id)
    )


def where_entry_type(entry_type):
    """ Generate a where clause for entries of certain types

    entry_type -- one or more entries to check against
    """
    if isinstance(entry_type, list):
        return model.Entry.entry_type << entry_type
    return model.Entry.entry_type == entry_type


def where_entry_type_not(entry_type):
    """ Generate a where clause for entries that aren't of certain types

    entry_type -- one or more entries to check against
    """
    if isinstance(entry_type, list):
        return model.Entry.entry_type.not_in(entry_type)
    return model.Entry.entry_type != entry_type


def where_entry_date(datespec):
    """ Where clause for entries which match a textual date spec

    datespec -- The date spec to check for, in YYYY[[-]MM[[-]DD]] format
    """
    date, interval, _ = utils.parse_date(datespec)
    start_date, end_date = date.span(interval)

    print('where_entry_date', start_date, end_date)

    return ((model.Entry.entry_date >= start_date.to('utc').datetime) &
            (model.Entry.entry_date <= end_date.to('utc').datetime))


def get_entry(entry):
    """ Helper function to get an entry by ID or by object """

    if hasattr(entry, 'id'):
        return entry

    if isinstance(entry, (int, str)):
        return model.Entry.get(model.Entry.id == int(entry))
    raise ValueError("entry is of unknown type {}".format(type(entry)))


def build_query(spec):
    """ build the where clause based on a view specification

    spec -- The view specification. Contains the following possible values:
        future -- Boolean; whether to include entries from the future
        category -- Which category to limit to
        recurse -- Whether to include subcategories
        entry_type -- one or more entry types to include
        entry_type_not -- one or more entry types to exclude
        date -- a date spec
        last -- the last entry to end a view on
        first -- the first entry to start a view on
        before -- get entries from before this one
        after -- get entries from after this one
    """

    # primarily restrict by publication status
    if spec.get('future', False):
        where = where_entry_visible_future()
    else:
        where = where_entry_visible()

    # restrict by category
    if 'category' in spec or 'recurse' in spec:
        path = str(spec.get('category', ''))
        recurse = spec.get('recurse', False)
        where = where & where_entry_category(path, recurse)

    if 'entry_type' in spec:
        where = where & where_entry_type(spec['entry_type'])

    if 'entry_type_not' in spec:
        where = where & where_entry_type_not(spec['entry_type_not'])

    if 'date' in spec:
        where = where & where_entry_date(spec['date'])

    if 'last' in spec:
        where = where & where_entry_last(get_entry(spec['last']))

    if 'first' in spec:
        where = where & where_entry_first(get_entry(spec['first']))

    if 'before' in spec:
        where = where & where_before_entry(get_entry(spec['before']))

    if 'after' in spec:
        where = where & where_after_entry(get_entry(spec['after']))

    return where



""" Publ entry point """

import time

import arrow
import flask

from . import config, rendering, model, index, caching, view, utils
from .caching import cache


def publ(name, cfg):
    """ Create a Flask app and configure it for use with Publ """

    config.setup(cfg)

    app = flask.Flask(name,
                      template_folder=config.template_folder,
                      static_folder=config.static_folder,
                      static_url_path=config.static_url_path)

    for route in [
            '/',
            '/<path:category>/',
            '/<template>',
            '/<path:category>/<template>',
    ]:
        app.add_url_rule(route, 'category', rendering.render_category)

    for route in [
            '/<int:entry_id>',
            '/<int:entry_id>-<slug_text>',
            '/<path:category>/<int:entry_id>',
            '/<path:category>/<int:entry_id>-<slug_text>',
    ]:
        app.add_url_rule(route, 'entry', rendering.render_entry)

    app.add_url_rule('/<path:path>.PUBL_PATHALIAS',
                     'path_alias', rendering.render_path_alias)

    if not app.debug:
        app.register_error_handler(Exception, rendering.render_exception)

    app.jinja_env.globals.update(  # pylint: disable=no-member
        get_view=view.get_view, arrow=arrow, static=utils.static_url)

    if config.index_rescan_interval:
        app.before_request(scan_index)

    if 'CACHE_THRESHOLD' in config.cache:
        app.after_request(set_cache_expiry)

    cache.init_app(app)

    # Scan the index
    model.setup()
    scan_index(True)
    index.background_scan(config.content_folder)

    return app


last_scan = None  # pylint: disable=invalid-name


def scan_index(force=False):
    """ Rescan the index if it's been more than a minute since the last scan """
    global last_scan  # pylint: disable=invalid-name,global-statement
    now = time.time()
    if force or not last_scan or now - last_scan > config.index_rescan_interval:
        index.scan_index(config.content_folder)
        last_scan = now


def set_cache_expiry(req):
    """ Set the cache control headers """
    if 'CACHE_THRESHOLD' in config.cache:
        req.headers['Cache-Control'] = (
            'public, max-age={}'.format(config.cache['CACHE_THRESHOLD']))
    return req



# view.py
""" A view of entries """

from __future__ import absolute_import, with_statement

import arrow
import flask

from . import model, utils, queries
from .entry import Entry

# Prioritization list for page/offset/whatever
OFFSET_PRIORITY = ['date', 'last', 'first', 'before', 'after']

# Prioritization list for pagination type
#
# NOTE: date appears in here too so that if it appears as an offset it
# overrides the count
PAGINATION_PRIORITY = ['date', 'count']


#: Ordering queries for different sort orders
ORDER_BY = {
    'newest': [-model.Entry.entry_date, -model.Entry.id],
    'oldest': [model.Entry.entry_date, model.Entry.id],
    'title': [model.Entry.title, model.Entry.id]
}

REVERSE_ORDER_BY = {
    'newest': [model.Entry.entry_date, model.Entry.id],
    'oldest': [-model.Entry.entry_date, -model.Entry.id],
    'title': [-model.Entry.title, -model.Entry.id]
}


class View:
    # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """ A view of entries """

    def __init__(self, input_spec=None):
        """ Generate a view.

        input_spec -- the parameters to the view. In addition to the values provided
        to queries.build_query, this also takes the following keys:
            count -- How many entries to include in a page
            order -- How to order the entries ('newest' or 'oldest')
        """

        # filter out any priority override things
        spec = {
            k: v for k, v in input_spec.items()
            if k not in [*OFFSET_PRIORITY, *PAGINATION_PRIORITY]
        }

        # pull in the first offset type that appears
        for offset in OFFSET_PRIORITY:
            if offset in input_spec:
                spec[offset] = input_spec[offset]
                break

        # pull in the first page type that appears
        for pagination in PAGINATION_PRIORITY:
            if pagination in input_spec:
                spec[pagination] = input_spec[pagination]
                break

        self.spec = spec
        self._where = queries.build_query(spec)
        self._query = model.Entry.select().where(self._where)

        if 'count' in spec:
            self._query = self._query.limit(spec['count'])

        self._order_by = spec.get('order', 'newest')
        self._entries = self._query.order_by(*ORDER_BY[self._order_by])

        self.link = utils.CallableProxy(self._link)

        if 'date' in self.spec:
            _, self.type, _ = utils.parse_date(self.spec['date'])
        elif 'count' in self.spec:
            self.type = 'count'
        else:
            self.type = None

    def _spec_filtered(self):
        # Return a version of our spec where all pagination stuff has been
        # removed
        return {k: v for k, v in self.spec.items()
                if k not in [*OFFSET_PRIORITY, *PAGINATION_PRIORITY]}

    def __str__(self):
        return str(self._link())

    def _link(self, template='', absolute=False):
        args = {}
        for k, val in self.spec.items():
            if k in ['date', 'last', 'first', 'before', 'after']:
                if isinstance(val, (str, int)):
                    args[k] = val
                elif hasattr(val, 'id'):
                    # the item was an object, so we want the object's id
                    args[k] = val.id
                else:
                    raise ValueError(
                        "key {} is of type {}".format(k, type(val)))

        return flask.url_for('category',
                             **args,
                             template=template,
                             category=self.spec.get('category'),
                             _external=absolute)

    @property
    def first(self):
        """ Gets the first entry in the view """
        return self.entries[0] if self.entries else None

    @property
    def last(self):
        """ Gets the last entry in the view """
        return self.entries[-1] if self.entries else None

    def __getattr__(self, name):
        """ Lazy evaluation of properties """
        # pylint: disable=attribute-defined-outside-init

        if name == 'entries':
            self.entries = [Entry(e) for e in self._entries]
            return self.entries

        if name == 'last_modified':
            # Get the most recent entry in the view
            try:
                latest = self._query.order_by(-model.Entry.entry_date)[0]
                self.last_modified = arrow.get(Entry(latest).last_modified)
            except IndexError:
                self.last_modified = arrow.get()
            return self.last_modified

        if name == 'previous' or name == 'next':
            self.previous, self.next = self._get_pagination()
            return getattr(self, name)

        if name == 'newest' or name == 'oldest':
            if self._order_by == 'newest':
                self.newest, self.oldest = self.first, self.last
            elif self._order_by == 'oldest':
                self.newest, self.oldest = self.last, self.first
            else:
                raise ValueError(
                    'newest/oldest not supported on sort type {}'.format(self._order_by))
            return getattr(self, name)

        raise AttributeError("Unknown view attribute {}".format(name))

    def _get_pagination(self):
        """ Compute the next/previous pages from this view.

        Returns a tuple of previous page, next page.
        """

        # https://github.com/fluffy-critter/Publ/issues/35
        oldest = self.oldest
        newest = self.newest
        oldest_neighbor = self.oldest.previous if oldest else None
        newest_neighbor = self.newest.next if newest else None

        base = self._spec_filtered()

        if 'count' in self.spec:
            return self._get_count_pagination(base, oldest_neighbor, newest_neighbor)

        if 'date' in self.spec:
            return self._get_date_pagination(base, oldest_neighbor, newest_neighbor)

        # we're not paginating?
        return None, None

    def _get_count_pagination(self, base, oldest_neighbor, newest_neighbor):
        """ Compute the pagination for count paginations """
        count = int(self.spec['count'])

        if self._order_by == 'newest':
                # Newest first; next page ends at the one prior to our oldest
            if oldest_neighbor:
                next_view = View({**base, 'last': oldest_neighbor})
            else:
                next_view = None

            # Previous page ends at [count] after our newest
            if newest_neighbor:
                # Ask for the next chunk of items in ascending order
                scan_view = View({**base,
                                  'first': newest_neighbor,
                                  'count': count,
                                  'order': 'oldest'})
                # our previous page starts at the last entry (ascending)
                previous_view = (View({**base, 'last': scan_view.entries[-1]})
                                 if scan_view.entries else None)
            else:
                previous_view = None

            return previous_view, next_view

        if self._order_by == 'oldest':
            # Oldest first; next page begins with our newest entry
            if newest_neighbor:
                next_view = View({**base, 'first': newest_neighbor})
            else:
                next_view = None

            # Previous page starts at [count] before our newest
            if oldest_neighbor:
                # Ask for the previous chunk of items in descending order
                scan_view = View({**base,
                                  'last': oldest_neighbor,
                                  'count': count,
                                  'order': 'newest'})
                # our previous page starts at the last entry (descending)
                previous_view = (View({**base, 'first': scan_view.entries[-1]})
                                 if scan_view.entries else None)
            else:
                previous_view = None

            return previous_view, next_view

        return None, None

    def _get_date_pagination(self, base, oldest_neighbor, newest_neighbor):
        """ Compute the pagination for date-based views """
        _, _, date_format = utils.parse_date(self.spec['date'])

        if self._order_by == 'newest':
                # newest first; next page contains the oldest neighbor
            next_date = oldest_neighbor.date.format(
                date_format) if oldest_neighbor else None
            previous_date = newest_neighbor.date.format(
                date_format) if newest_neighbor else None
        elif self._order_by == 'oldest':
            # oldest first; next page contains the newest neighbor
            next_date = newest_neighbor.date.format(
                date_format) if newest_neighbor else None
            previous_date = oldest_neighbor.date.format(
                date_format) if oldest_neighbor else None
        else:
            raise ValueError(
                "Unsupported sort {} for date pagination".format(self._order_by))

        previous_view = View({**base, 'date': previous_date}) if previous_date else None
        next_view = View({**base, 'date': next_date}) if next_date else None

        return previous_view, next_view

    def __call__(self, **restrict):
        return View({**self.spec, **restrict})


def get_view(**kwargs):
    """ Wrapper function for constructing a view from scratch """
    return View(input_spec=kwargs)



# path_alias.py
""" Handling for URL aliases """

from __future__ import absolute_import, with_statement

from flask import url_for

from . import model


def set_alias(path, entry=None, url=None):
    """ Set a path alias.

    Arguments:

    path -- The path to alias
    entry -- The entry to alias it to
    url -- The external URL to alias it to
    """

    record, created = model.PathAlias.get_or_create(
        path=path,
        defaults={
            'redirect_entry': entry,
            'redirect_url': url
        })
    if not created:
        record.redirect_entry = entry
        record.redirect_url = url
        record.save()


def remove_alias(path):
    """ Remove a path alias.

    Arguments:

    path -- the path to remove the alias of
    """
    model.PathAlias.delete().where(model.PathAlias.path == path).execute()


def get_redirect(paths):
    """ Get a redirect from a path or list of paths

    Arguments:

    paths -- either a single path string, or a list of paths to check
    """

    if isinstance(paths, str):
        paths = [paths]

    for path in paths:
        record = model.PathAlias.get_or_none(model.PathAlias.path == path)
        if record and record.redirect_entry:
            return url_for('entry',
                           entry_id=record.redirect_entry.id,
                           category=record.redirect_entry.category,
                           slug_text=record.redirect_entry.slug_text)
        elif record and record.redirect_url:
            return record.redirect_url
    return None



# model.py
""" Database schema for the content index """

from __future__ import absolute_import

import logging
import threading
from enum import Enum

import peewee
from peewee import Proxy, Model, IntegerField, DateTimeField, CharField
import playhouse.db_url

from . import config

DATABASE_PROXY = Proxy()
lock = threading.Lock()  # pylint: disable=invalid-name

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Schema version; bump this whenever an existing table changes
SCHEMA_VERSION = 6


class BaseModel(Model):
    """ Base model for our content index """
    # pylint: disable=too-few-public-methods

    class Meta:
        """ database configuration """
        database = DATABASE_PROXY


class PublishStatus(Enum):
    """ The status of the entry """
    DRAFT = 0  # Entry should not be rendered
    HIDDEN = 1  # Entry should be shown via direct link, but not shown on a view
    PUBLISHED = 2  # Entry is visible
    SCHEDULED = 3  # Entry will be visible in the future

    @staticmethod
    class Field(IntegerField):
        """ Database mapping for PublishStatis """

        def db_value(self, value):
            """ map enum to database column """
            return value.value

        def python_value(self, value):
            """ map database column to enum """
            return PublishStatus(value)


class Global(BaseModel):
    """ key-value storage for the index itself """
    key = CharField(unique=True)
    int_value = IntegerField(null=True)
    str_value = CharField(null=True)


class FileMTime(BaseModel):
    """ File modification time """
    file_path = CharField(unique=True)
    stat_mtime = IntegerField()  # At least on SQLite, this is Y2k38-ready


class Entry(BaseModel):
    """ Indexed entry """
    file_path = CharField()
    category = CharField()
    status = PublishStatus.Field()
    entry_date = DateTimeField()  # UTC-normalized, for queries
    display_date = DateTimeField()  # arbitrary timezone, for display
    slug_text = CharField()
    entry_type = CharField()
    redirect_url = CharField(null=True)
    title = CharField(null=True)

    class Meta:
        """ meta info """
        # pylint: disable=too-few-public-methods
        indexes = (
            (('category', 'entry_type', 'entry_date'), False),
        )


class PathAlias(BaseModel):
    """ Path alias mapping """
    path = CharField(unique=True)
    redirect_url = CharField(null=True)
    redirect_entry = peewee.ForeignKeyField(
        Entry,
        null=True,
        backref='aliases')


class Image(BaseModel):
    """ Image metadata """
    file_path = CharField(unique=True)
    checksum = CharField()
    width = IntegerField()
    height = IntegerField()
    transparent = peewee.BooleanField()
    mtime = IntegerField()

ALL_TYPES = [
    Global,
    FileMTime,
    Entry,
    PathAlias,
    Image,
]


def setup():
    """ Set up the database """
    database = playhouse.db_url.connect(config.database)
    DATABASE_PROXY.initialize(database)

    rebuild = False
    try:
        cur_version = Global.get(key='schema_version').int_value
        logger.info("Current schema version: %s", cur_version)
        rebuild = cur_version != SCHEMA_VERSION
    except:  # pylint: disable=bare-except
        logger.info("Schema information not found")
        rebuild = True

    if rebuild:
        logger.info("Updating database schema")
        database.drop_tables(ALL_TYPES)

    database.create_tables(ALL_TYPES)

    version_record, _ = Global.get_or_create(key='schema_version')
    version_record.int_value = SCHEMA_VERSION
    version_record.save()



# image.py
''' Managing image renditions '''

from __future__ import absolute_import, with_statement

import os
import hashlib
import logging

import PIL.Image

from . import config
from . import model, utils

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Image:
    """ The basic Image class, which knows about the base version and how to
    generate renditions from it """

    def __init__(self, record):
        """ Get the base image from an index record """
        self._record = record

    def get_rendition(self, output_scale, kwargs):
        """
        Get the rendition for this image, generating it if necessary.
        Returns a tuple of `(relative_path, width, height)`, where relative_path
        is relative to the static file directory (i.e. what one would pass into
        `get_static()`)

        output_scale -- the upsample factor for the requested rendition

        Keyword arguments:

        scale -- the downsample factor for the base rendition
        scale_min_width -- the minimum width after downsampling
        scale_min_height -- the minimum height after downsampling
        width -- the width to target
        height -- the height to target
        max_width -- the maximum width
        max_height -- the maximum height
        resize -- how to fit the width and height; "fit", "fill", or "stretch"
        fill_crop_x -- horizontal offset fraction for resize="fill"
        fill_crop_y -- vertical offset fraction for resize="fill"
        format -- output format
        background -- background color when converting transparent to opaque
        quality -- the JPEG quality to save the image as
        """

        # pylint:disable=too-many-locals

        input_filename = self._record.file_path
        basename, ext = os.path.splitext(os.path.basename(input_filename))
        basename = utils.make_slug(basename)

        if kwargs.get('format'):
            ext = '.' + kwargs['format']

        # The spec for building the output filename
        out_spec = [basename, self._record.checksum[-10:]]

        out_args = {}
        if ext in ['.png', '.jpg', '.jpeg']:
            out_args['optimize'] = True

        size, box = self.get_rendition_size(kwargs, output_scale)
        if size and (size[0] < self._record.width or size[1] < self._record.height):
            out_spec.append('x'.join([str(v) for v in size]))
        if box:
            out_spec.append('-'.join([str(v) for v in box]))

        # Set RGBA flattening options
        flatten = self._record.transparent and ext not in ['.png', '.gif']
        if flatten and 'background' in kwargs:
            bg_color = kwargs['background']
            if isinstance(bg_color, (tuple, list)):
                out_spec.append('b' + '-'.join([str(a) for a in bg_color]))
            else:
                out_spec.append('b' + str(bg_color))

        # Set JPEG quality
        if (ext == '.jpg' or ext == '.jpeg') and kwargs.get('quality'):
            out_spec.append('q' + str(kwargs['quality']))
            out_args['quality'] = kwargs['quality']

        # Build the output filename
        out_basename = '_'.join([str(s) for s in out_spec]) + ext
        out_rel_path = os.path.join(
            config.image_output_subdir,
            self._record.checksum[0:2],
            self._record.checksum[2:6],
            out_basename)
        out_fullpath = os.path.join(config.static_folder, out_rel_path)

        if not os.path.isfile(out_fullpath):
            logger.info("Rendering file %s", out_fullpath)
            if not os.path.isdir(os.path.dirname(out_fullpath)):
                os.makedirs(os.path.dirname(out_fullpath))

            image = PIL.Image.open(input_filename)

            if size:
                image = image.resize(size=size, box=box,
                                     resample=PIL.Image.LANCZOS)
            if flatten:
                image = self.flatten(image, kwargs.get('background'))
            image.save(out_fullpath, **out_args)

        return out_rel_path, size

    def get_rendition_size(self, spec, output_scale):
        """
        Wrapper to determine the overall rendition size and cropping box

        Returns tuple of (size,box)
        """

        mode = spec.get('resize', 'fit')

        if mode == 'fit':
            return self.get_rendition_fit_size(spec, output_scale)

        if mode == 'fill':
            return self.get_rendition_fill_size(spec, output_scale)

        if mode == 'stretch':
            return self.get_rendition_stretch_size(spec, output_scale)

        raise ValueError("Unknown resize mode {}".format(mode))

    def get_rendition_fit_size(self, spec, output_scale):
        """ Determine the scaled size based on the provided spec """

        input_w = self._record.width  # input width
        input_h = self._record.height  # input height

        width = input_w
        height = input_h

        scale = spec.get('scale')
        if scale:
            width = width / scale
            height = height / scale

        min_width = spec.get('scale_min_width')
        if min_width and width < min_width:
            height = height * min_width / width
            width = min_width

        min_height = spec.get('scale_min_height')
        if min_height and height < min_height:
            width = width * min_height / height
            height = min_height

        tgt_width, tgt_height = spec.get('width'), spec.get('height')

        if tgt_width and width > tgt_width:
            height = height * tgt_width / width
            width = tgt_width

        tgt_height = spec.get('height')
        if tgt_height and height > tgt_height:
            width = width * tgt_height / height
            height = tgt_height

        tgt_width, tgt_height = spec.get('max_width'), spec.get('max_height')

        if tgt_width and width > tgt_width:
            height = height * tgt_width / width
            width = tgt_width

        tgt_height = spec.get('height')
        if tgt_height and height > tgt_height:
            width = width * tgt_height / height
            height = tgt_height

        width = width * output_scale
        height = height * output_scale

        # Never scale to larger than the base rendition
        width = min(round(width), input_w)
        height = min(round(height), input_h)

        return (width, height), None

    def get_rendition_fill_size(self, spec, output_scale):
        """ Determine the scale-crop size given the provided spec """

        input_w = self._record.width
        input_h = self._record.height

        width = input_w
        height = input_h

        scale = spec.get('scale')
        if scale:
            width = width / scale
            height = height / scale

        if spec.get('scale_min_width'):
            width = max(width, spec['spec_min_width'])

        if spec.get('scale_min_height'):
            height = max(height, spec['scale_min_height'])

        if spec.get('width'):
            width = min(width, spec['width'])
        if spec.get('max_width'):
            width = min(width, spec['max_width'])

        if spec.get('height'):
            height = min(height, spec['height'])
        if spec.get('max_height'):
            height = min(height, spec['max_height'])

        width = width * output_scale
        height = height * output_scale

        # Never scale to larger than the base rendition (but keep the output
        # aspect)
        if width > input_w:
            height = height * input_w / width
            width = input_w

        if height > input_h:
            width = width * input_h / height
            height = input_h

        # Determine the box size
        box_w = min(input_w, round(width * input_h / height))
        box_h = min(input_h, round(height * input_w / width))

        # Box offset
        box_x = round((input_w - box_w) * spec.get('fill_crop_x', 0.5))
        box_y = round((input_h - box_h) * spec.get('fill_crop_y', 0.5))

        return (round(width), round(height)), (box_x, box_y, box_x + box_w, box_y + box_h)

    def get_rendition_stretch_size(self, spec, output_scale):
        """ Determine the scale-crop size given the provided spec """

        width = self._record.width
        height = self._record.height

        scale = spec.get('scale')
        if scale:
            width = width / scale
            height = height / scale

        min_width = spec.get('scale_min_width')
        if min_width and width < min_width:
            width = min_width

        min_height = spec.get('scale_min_height')
        if min_height and height < min_height:
            height = min_height

        tgt_width, tgt_height = spec.get('width'), spec.get('height')

        if tgt_width and width > tgt_width:
            width = tgt_width

        tgt_height = spec.get('height')
        if tgt_height and height > tgt_height:
            height = tgt_height

        tgt_width, tgt_height = spec.get('max_width'), spec.get('max_height')

        if tgt_width and width > tgt_width:
            width = tgt_width

        tgt_height = spec.get('height')
        if tgt_height and height > tgt_height:
            height = tgt_height

        width = width * output_scale
        height = height * output_scale

        return (round(width), round(height)), None

    @staticmethod
    def flatten(image, bgcolor=None):
        """ Flatten an image, with an optional background color """
        if bgcolor:
            background = PIL.Image.new('RGB', image.size, bgcolor)
            background.paste(image, mask=image.split()[3])
            return background

        return image.convert('RGB')


def get_image(path, search_path):
    """ Get an Image object. Arguments:

    path -- the image's filename
    search_path -- a search path or list of search paths
    """

    file_path = utils.find_file(path, search_path)
    if not file_path:
        return None

    record = model.Image.get_or_none(file_path=file_path)
    mtime = os.stat(file_path).st_mtime
    if not record or record.mtime < mtime:
        # Reindex the file
        logger.info("Updating image %s", file_path)

        # compute the md5sum; from https://stackoverflow.com/a/3431838/318857
        md5 = hashlib.md5()
        with open(file_path, 'rb') as file:
            for chunk in iter(lambda: file.read(16384), b""):
                md5.update(chunk)

        image = PIL.Image.open(file_path)
        values = {
            'checksum': md5.hexdigest(),
            'width': image.width,
            'height': image.height,
            'mtime': mtime,
            'transparent': image.mode == 'RGBA'
        }
        record, created = model.Image.get_or_create(
            file_path=file_path, defaults=values)
        if not created:
            record.update(**values).where(model.Image.id == record.id)

    return Image(record)



# template.py
""" Wrapper for template information """

from __future__ import absolute_import, with_statement

import os
import arrow


class Template:
    """ Template information wrapper """
    # pylint: disable=too-few-public-methods

    def __init__(self, name, filename, file_path):
        """ Useful information for the template object:

        name -- The name of the template
        filename -- The filename of the template
        file_path -- The full path to the template
        """
        self.name = name
        self.filename = filename
        self.last_modified = arrow.get(os.stat(file_path).st_mtime)

    def __str__(self):
        return self.name



# item.py
""" Functions for handling content items """

from __future__ import absolute_import, with_statement

import email
import functools
import logging
import os
import random
import re
import shutil
import tempfile
import uuid

import arrow
import flask

from . import config
from . import model
from . import queries
from . import path_alias
from . import markdown
from . import utils
from .utils import CallableProxy, TrueCallableProxy, make_slug

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@functools.lru_cache(10)
def load_message(filepath):
    """ Load a message from the filesystem """
    with open(filepath, 'r') as file:
        return email.message_from_file(file)


class Entry:
    """ A wrapper for an entry. Lazily loads the actual message data when
    necessary.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, record):
        """ Construct an Entry wrapper

        record -- the index record to use as the basis
        """

        self._record = record   # index record
        self._message = None    # actual message payload, lazy-loaded

        self.date = arrow.get(record.display_date)

        self.link = CallableProxy(self._link)
        self.permalink = CallableProxy(self._permalink)
        self.archive = CallableProxy(self._archive_link)

        self.next = CallableProxy(self._next)
        self.previous = CallableProxy(self._previous)

        self._relative_search_path = [
            os.path.dirname(self._record.file_path),
            os.path.join(config.content_folder, self._record.category),
            config.content_folder,
            config.static_folder,
        ]
        self._absolute_search_path = [
            config.content_folder,
            config.static_folder,
        ]

    def _link(self, *args, **kwargs):
        """ Returns a link, potentially pre-redirected """
        if self._record.redirect_url:
            return self._record.redirect_url

        return self._permalink(*args, **kwargs)

    def _permalink(self, absolute=False, expand=True):
        """ Returns a canonical URL for the item """
        return flask.url_for('entry',
                             entry_id=self._record.id,
                             category=self._record.category if expand else None,
                             slug_text=self._record.slug_text if expand else None,
                             _external=absolute)

    def _archive_link(self, paging=None, template='', category=None, absolute=False):
        args = {
            'template': template,
            'category': category if category is not None else self.category,
        }
        if paging == 'day':
            args['date'] = self.date.format(utils.DAY_FORMAT)
        elif paging == 'month':
            args['date'] = self.date.format(utils.MONTH_FORMAT)
        elif paging == 'year':
            args['date'] = self.date.format(utils.YEAR_FORMAT)
        else:
            args['first'] = self._record.id

        return flask.url_for('category', **args, _external=absolute)

    def _load(self):
        """ ensure the message payload is loaded """
        # pylint: disable=attribute-defined-outside-init

        if not self._message:
            filepath = self._record.file_path
            try:
                self._message = load_message(filepath)
            except FileNotFoundError:
                expire_record(self._record)

            body, _, more = self._message.get_payload().partition('\n.....\n')
            if not more and body.startswith('.....\n'):
                # The entry began with a cut, which failed to parse.
                # This rule is easier/faster than dealing with a regex from
                # hell.
                more = body[6:]
                body = ''

            _, ext = os.path.splitext(filepath)
            is_markdown = ext == '.md'
            self.body = TrueCallableProxy(
                self._get_markup,
                body,
                is_markdown) if body else CallableProxy(lambda **kwargs: '')
            self.more = TrueCallableProxy(
                self._get_markup,
                more,
                is_markdown) if more else CallableProxy(lambda **kwargs: '')

            self.last_modified = arrow.get(
                os.stat(self._record.file_path).st_mtime).to(config.timezone)

            return True
        return False

    def _get_markup(self, text, is_markdown, **kwargs):
        """ get the rendered markup for an entry

            is_markdown -- whether the entry is formatted as Markdown
            kwargs -- parameters to pass to the Markdown processor
        """
        if is_markdown:
            md_config = {
                **kwargs,
                "relative_search_path": self._relative_search_path,
                "absolute_search_path": self._absolute_search_path
            }

            return flask.Markup(markdown.to_html(text, config=md_config))
        return flask.Markup(text)

    def __getattr__(self, name):
        """ Lazy binding for deferred properties """
        # pylint: disable=attribute-defined-outside-init
        if name == 'previous':
            # Get the previous entry in the same category (by date)
            self.previous = self.previous_in(self._record.category, False)
            return self.previous

        if name == 'next':
            # Get the next entry in the same category (by date)
            self.next = self.next_in(self._record.category, False)
            return self.next

        if hasattr(self._record, name):
            return getattr(self._record, name)

        if self._load():
            # We just loaded which modifies our own attrs, so rerun the default
            # logic
            return getattr(self, name)

        return self._message.get(name)

    @staticmethod
    def _get_first(query):
        """ Get the first entry in a query result """
        query = query.limit(1)
        return Entry(query[0]) if query.count() else None

    def _previous(self, **kwargs):
        # Get the previous item in any particular category
        spec = {
            'category': self._record.category,
            'recurse': 'category' in kwargs
        }
        spec.update(kwargs)

        return self._get_first(model.Entry.select().where(
            queries.build_query(spec) &
            queries.where_before_entry(self._record)
        ).order_by(-model.Entry.entry_date, -model.Entry.id))

    def _next(self, **kwargs):
        # Get the next item in any particular category
        spec = {
            'category': self._record.category,
            'recurse': 'category' in kwargs
        }
        spec.update(kwargs)

        return self._get_first(
            model.Entry.select().where(
                queries.build_query(spec) &
                queries.where_after_entry(self._record)
            ).order_by(model.Entry.entry_date, model.Entry.id))

    def get(self, name):
        """ Get a single header on an entry """

        self._load()
        return self._message.get(name)

    def get_all(self, name):
        """ Get all related headers on an entry, as an iterable list """
        self._load()
        return self._message.get_all(name) or []


def guess_title(basename):
    """ Attempt to guess the title from the filename """

    base, _ = os.path.splitext(basename)
    return re.sub(r'[ _-]+', r' ', base).title()


def get_entry_id(entry, fullpath, assign_id):
    """ Get or generate an entry ID for an entry """
    warn_duplicate = False

    if 'Entry-ID' in entry:
        entry_id = int(entry['Entry-ID'])
    else:
        entry_id = None

    # See if we've inadvertently duplicated an entry ID
    if entry_id:
        other_entry = model.Entry.get_or_none(model.Entry.id == entry_id)
        if (other_entry
                and other_entry.file_path != fullpath
                and os.path.isfile(other_entry.file_path)):
            warn_duplicate = entry_id
            entry_id = None

    # Do we need to assign a new ID?
    if not entry_id and not assign_id:
        # We're not assigning IDs yet
        return None

    if not entry_id:
        # See if we already have an entry with this file path
        by_filepath = model.Entry.get_or_none(file_path=fullpath)
        if by_filepath:
            entry_id = by_filepath.id

    if not entry_id:
        # We still don't have an ID; generate one randomly. Experiments find that this approach
        # averages around 0.25 collisions per ID generated while keeping the
        # entry ID reasonably short. count*N+C averages 1/(N-1) collisions
        # per ID.

        # database=None is to shut up pylint
        limit = max(10, model.Entry.select().count(database=None) * 5)

        entry_id = random.randint(1, limit)
        while model.Entry.get_or_none(model.Entry.id == entry_id):
            entry_id = random.randint(1, limit)

    if warn_duplicate is not False:
        logger.warning("Entry '%s' had ID %d, which belongs to '%s'. Reassigned to %d",
                       fullpath, warn_duplicate, other_entry.file_path, entry_id)

    return entry_id


def scan_file(fullpath, relpath, assign_id):
    """ scan a file and put it into the index """

    # Since a file has changed, the lrucache is invalid.
    load_message.cache_clear()

    try:
        entry = load_message(fullpath)
    except FileNotFoundError:
        # The file doesn't exist, so remove it from the index
        record = model.Entry.get_or_none(file_path=fullpath)
        if record:
            expire_record(record)
        return True

    with model.lock:
        entry_id = get_entry_id(entry, fullpath, assign_id)
        if entry_id is None:
            return False

        fixup_needed = (str(entry_id) != entry.get('Entry-ID')
                        or 'Date' not in entry
                        or 'UUID' not in entry)

        basename = os.path.basename(relpath)
        title = entry['title'] or guess_title(basename)

        values = {
            'file_path': fullpath,
            'category': entry.get('Category', os.path.dirname(relpath)),
            'status': model.PublishStatus[entry.get('Status', 'SCHEDULED').upper()],
            'entry_type': entry.get('Entry-Type', ''),
            'slug_text': make_slug(entry['Slug-Text'] or title),
            'redirect_url': entry['Redirect-To'],
            'title': title,
        }

        if 'Date' in entry:
            entry_date = arrow.get(entry['Date'], tzinfo=config.timezone)
            del entry['Date']
            entry['Date'] = entry_date.format()
        else:
            entry_date = arrow.get(
                os.stat(fullpath).st_ctime).to(config.timezone)
            entry['Date'] = entry_date.format()

        values['entry_date'] = entry_date.to('utc').datetime
        values['display_date'] = entry_date.datetime

        logger.debug("getting entry %s with id %d", fullpath, entry_id)
        record, created = model.Entry.get_or_create(
            id=entry_id, defaults=values)

        if not created:
            logger.debug("Reusing existing entry %d", record.id)
            record.update(**values).where(model.Entry.id ==
                                          record.id).execute()

        # Update the entry ID
        del entry['Entry-ID']
        entry['Entry-ID'] = str(record.id)

        if not 'UUID' in entry:
            entry['UUID'] = str(uuid.uuid4())

        # add other relationships to the index
        for alias in entry.get_all('Path-Alias', []):
            path_alias.set_alias(alias, entry=record)
        for alias in entry.get_all('Path-Unalias', []):
            path_alias.remove_alias(alias)

        if fixup_needed:
            with tempfile.NamedTemporaryFile('w', delete=False) as file:
                tmpfile = file.name
                file.write(str(entry))
            shutil.move(tmpfile, fullpath)

        return record


def expire_record(record):
    """ Expire a record for a missing entry """
    load_message.cache_clear()

    with model.lock:
        # This entry no longer exists so delete it, and anything that references it
        # SQLite doesn't support cascading deletes so let's just clean up
        # manually
        model.PathAlias.delete().where(model.PathAlias.redirect_entry == record).execute()
        record.delete_instance(recursive=True)



# markdown.py
""" handler for markdown formatting """

from __future__ import absolute_import

import re
import ast
import os
import logging

import misaka
import flask

import pygments
import pygments.formatters
import pygments.lexers

from . import image, utils

ENABLED_EXTENSIONS = [
    'fenced-code', 'footnotes', 'strikethrough', 'highlight', 'math', 'math-explicit'
]

CSS_SIZE_MODE = {
    'fit': 'contain',
    'fill': 'cover'
}

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class HtmlRenderer(misaka.HtmlRenderer):
    """ Customized renderer for enhancing Markdown formatting """

    def __init__(self, config):
        super().__init__()
        self._config = config
        self._relative_search_path = config.get(
            'relative_search_path', config.get('search_path'))
        self._absolute_search_path = config.get(
            'absolute_search_path', config.get('search_path'))

    def image(self, raw_url, title='', alt=''):
        """ Adapt a standard Markdown image to a generated rendition """
        # pylint: disable=too-many-locals

        text = ''

        image_specs = raw_url
        if title:
            image_specs += ' "{}"'.format(title)

        alt, container_args = self._parse_alt_text(alt)

        spec_list = [spec.strip() for spec in image_specs.split('|')]

        container_args = {**self._config, **container_args}

        if 'count' in container_args:
            if 'count_offset' in container_args:
                spec_list = spec_list[container_args['count_offset']:]
            spec_list = spec_list[:container_args['count']]

        if 'div_class' in container_args:
            text += self._make_tag('div',
                                   {'class': container_args['div_class']})

        for spec in spec_list:
            if not spec:
                continue

            text += self._render_image(spec,
                                       container_args,
                                       alt) if spec else ''

        if 'div_class' in container_args:
            text += '</div>'

        return text or ' '

    def blockcode(self, text, lang):
        """ Pass a code fence through pygments """
        if lang and self._config.get('highlight_syntax', 'True'):
            try:
                lexer = pygments.lexers.get_lexer_by_name(lang, stripall=True)
            except pygments.lexers.ClassNotFound:
                lexer = None

            if lexer:
                formatter = pygments.formatters.HtmlFormatter()  # pylint: disable=no-member
                return pygments.highlight(text, lexer, formatter)

        return '\n<div class="highlight"><pre>{}</pre></div>\n'.format(
            flask.escape(text.strip()))

    def _render_image(self, spec, container_args, alt_text=None):
        """ Render an image specification into an <img> tag """

        try:
            path, image_args, title = self._parse_image_spec(spec)
            composite_args = {**container_args, **image_args}

            if path.startswith('//') or '://' in path:
                return self._remote_image(path, composite_args, title, alt_text)

            return self._local_image(path, composite_args, title, alt_text)
        except Exception as err:  # pylint: disable=broad-except
            logger.exception("Got error on spec %s: %s", spec, err)
            return ('<span class="error">Couldn\'t parse image spec: ' +
                    '<code>{}</code> {}</span>'.format(flask.escape(spec),
                                                       flask.escape(str(err))))

    @staticmethod
    def _make_tag(name, attrs):
        text = '<' + name
        for key, val in attrs.items():
            if val is not None:
                text += ' {}="{}"'.format(key, flask.escape(val))
        text += '>'
        return text

    @staticmethod
    def _rendition_args(image_args, remap):
        """ Generate rendition arguments specific to a rendition. The 'remap'
        dict maps from destination key -> priority list of source keys
        """
        out_args = image_args
        for dest_key, src_keys in remap.items():
            remap_value = None
            if isinstance(src_keys, str):
                src_keys = [src_keys]

            for key in src_keys:
                if key in image_args:
                    remap_value = image_args[key]
                    break

            if remap_value is not None:
                if out_args is image_args:
                    out_args = {**image_args}
                out_args[dest_key] = remap_value

        return out_args

    def _local_image(self, path, image_args, title, alt_text):
        """ Render an img tag for a locally-stored image """

        # Determine the appropriate search path for the image request
        if os.path.isabs(path):
            search_path = self._absolute_search_path
            path = os.path.relpath(path, "/")
        else:
            search_path = self._relative_search_path

        # Get the image object
        img = image.get_image(path, search_path)
        if not img:
            return ('<span class="error">Couldn\'t find image: ' +
                    '<code>{}</code></span>'.format(flask.escape(path)))

        # Get the 1x and 2x renditions
        img_1x, size = img.get_rendition(
            1, self._rendition_args(image_args, {"quality": "quality_ldpi"}))
        img_2x, _ = img.get_rendition(
            2, self._rendition_args(image_args, {"quality": "quality_hdpi"}))

        # ... and their URLs
        absolute = self._config.get('absolute')
        img_1x = utils.static_url(img_1x, absolute)
        img_2x = utils.static_url(img_2x, absolute)

        text = self._make_tag('img', {
            'src': img_1x,
            'width': size[0],
            'height': size[1],
            'srcset': "{} 1x, {} 2x".format(img_1x, img_2x) if img_1x != img_2x else None,
            'title': title,
            'alt': alt_text
        })

        # Wrap it in a link as appropriate
        if 'link' in image_args and image_args['link'] is not None:
            text = '<a href="{}">{}</a>'.format(
                flask.escape(image_args['link']), text)
        elif 'gallery_id' in image_args and image_args['gallery_id'] is not None:
            text = '{}{}</a>'.format(
                self._fullsize_link(
                    img, image_args, title, absolute),
                text)

        return text

    def _fullsize_link(self, img, image_args, title, absolute):
        fullsize_args = {}
        for key in ['width', 'height', 'quality', 'format', 'background']:
            fsk = 'fullsize_' + key
            if fsk in image_args:
                fullsize_args[key] = image_args[fsk]

        img_fullsize, _ = img.get_rendition(1, fullsize_args)
        img_fullsize = utils.static_url(img_fullsize, absolute)

        return self._make_tag('a', {
            'href': img_fullsize,
            'data-lightbox': image_args['gallery_id'],
            'title': title
        })

    def _remote_image(self, path, image_args, title, alt_text):
        """ Render an img tag for a remotely-stored image """

        attrs = {
            'title': title,
            'alt': alt_text
        }

        # try to fudge the sizing
        width = image_args.get('width')
        height = image_args.get('height')
        size_mode = image_args.get('resize', 'fit')

        if width and height and size_mode != 'stretch':
            attrs['style'] = ';'.join([
                'background-image:url(\'{}\')'.format(flask.escape(path)),
                'background-size:{}'.format(CSS_SIZE_MODE[size_mode]),
                'background-position:{:.1f}% {:.1f}%'.format(
                    image_args.get('fill_crop_x', 0.5) * 100,
                    image_args.get('fill_crop_y', 0.5) * 100),
                'background-repeat:no-repeat'
            ])
            attrs['src'] = (
                'data:image/gif;base64,R0lGODlhAQABAAAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw'
            )
        else:
            attrs['src'] = path

        attrs['width'] = width
        attrs['height'] = height

        text = self._make_tag('img', attrs)

        if 'link' in image_args and image_args['link'] is not None:
            text = '<a href="{}">{}</a>'.format(
                flask.escape(image_args['link']), text)
        elif 'gallery_id' in image_args and image_args['gallery_id'] is not None:
            text = '{}{}</a>'.format(
                self._make_tag('a', {
                    'href': path,
                    'data-lightbox': image_args['gallery_id'],
                    'title': title
                }),
                text)

        return text

    def _parse_image_spec(self, spec):
        """ Parse an image spec out into (path,args,title) """

        # I was having trouble coming up with a single RE that did it right,
        # so let's just break it down into sub-problems. First, parse out the
        # alt text...
        match = re.match(r'(.+)\s+\"(.*)\"\s*$', spec)
        if match:
            spec, title = match.group(1, 2)
        else:
            title = None

        # and now parse out the arglist
        match = re.match(r'([^\{]*)(\{(.*)\})\s*$', spec)
        if match:
            spec = match.group(1)
            args = self._parse_args(match.group(3))
        else:
            args = {}

        return spec, args, title

    def _parse_alt_text(self, spec):
        """ Parse the alt text out into (alt_text,args) """
        match = re.match(r'([^\{]*)(\{(.*)\})$', spec)
        if match:
            spec = match.group(1)
            args = self._parse_args(match.group(3))
        else:
            args = {}

        return spec, args

    @staticmethod
    def _parse_args(args):
        """ Parse an arglist into args and kwargs """
        # per https://stackoverflow.com/a/49723227/318857

        args = 'f({})'.format(args)
        tree = ast.parse(args)
        funccall = tree.body[0].value

        args = [ast.literal_eval(arg) for arg in funccall.args]
        kwargs = {arg.arg: ast.literal_eval(arg.value)
                  for arg in funccall.keywords}

        if len(args) > 2:
            raise TypeError(
                "Expected at most 2 positional args but {} were given".format(len(args)))

        if len(args) >= 1:
            kwargs['width'] = int(args[0])
        if len(args) >= 2:
            kwargs['height'] = int(args[1])

        return kwargs


def to_html(text, config):
    """ Convert Markdown text to HTML """

    processor = misaka.Markdown(HtmlRenderer(config),
                                extensions=ENABLED_EXTENSIONS)
    return processor(text)



# config.py
""" configuration container for Publ """

import sys
from dateutil import tz

# pylint: disable=invalid-name

database = 'sqlite:///:memory:'
content_folder = 'content'
template_folder = 'templates'
static_folder = 'static'
static_url_path = '/static'
image_output_subdir = '_img'
index_rescan_interval = 300
timezone = tz.tzlocal()
cache = {}


def setup(cfg):
    """ set up the global configuration from an object """

    # copy the necessary configuration values over
    this_module = sys.modules[__name__]
    for name, value in cfg.items():
        if hasattr(this_module, name):
            setattr(this_module, name, value)



"""Setup for Publ packaging"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='Publ',

    version='0.1.7',

    description='A content-management system for flexible web-based publishing',

    long_description=long_description,

    long_description_content_type='text/markdown',

    url='https://github.com/fluffy-critter/Publ',

    author='fluffy',
    author_email='fluffy@beesbuzz.biz',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',

        'Framework :: Flask',

        'License :: OSI Approved :: MIT License',

        'Natural Language :: English',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: News/Diary',

    ],

    keywords='website cms publishing blog photogallery sharing',

    packages=['publ'],

    install_requires=[
        'Flask',
        'Flask-Caching',
        'arrow',
        'peewee',
        'misaka',
        'Pillow',
        'pygments',
        'watchdog',
    ],

    extras_require={
        'dev': ['pylint'],
    },

    project_urls={
        'Main Site': 'http://publ.beesbuzz.biz',
        'Bug Reports': 'https://github.com/fluffy-critter/Publ/issues',
        'Funding': 'https://patreon.com/fluffy',
        'Source': 'https://github.com/fluffy-critter/Publ/',
    },
)



# caching.py
""" Useful caching functions """

from flask_caching import Cache
from flask import request

from . import config

cache = Cache(config=config.cache)  # pylint: disable=invalid-name


def make_category_key():
    """ Key generator for categories """
    return 'category/' + request.full_path


def make_entry_key():
    """ Key generator for entries """
    return 'entry/' + request.path



# utils.py
""" Some useful utilities that don't belong anywhere else """

from __future__ import absolute_import, with_statement

import re
import os

import arrow
import flask

from . import config


class CallableProxy:
    """ Wrapper class to make args possible on properties """
    # pylint: disable=too-few-public-methods

    def __init__(self, func, *args, **kwargs):
        """ Construct the property proxy.

        func -- The function to wrap
        args -- Default positional arguments for the function call
        kwargs -- Default keyword arguments for the function call
        """

        self._func = func
        self._has_default = False
        self._default_args = args
        self._default_kwargs = kwargs

    def _get_default(self):
        """ Get the default function return """

        # pylint: disable=attribute-defined-outside-init
        if not self._has_default:
            if self._default_args:
                self._default = self._func(
                    *self._default_args,
                    **self._default_kwargs)
            else:
                self._default = self._func(**self._default_kwargs)
            self._has_default = True
        return self._default

    def __call__(self, *args, **kwargs):
        # use the new kwargs to override the defaults
        kwargs = dict(self._default_kwargs, **kwargs)

        # override args as well
        pos_args = [*args, *self._default_args[len(args):]]

        return self._func(*pos_args, **kwargs)

    def __getattr__(self, name):
        return getattr(self._get_default(), name)

    def __nonzero__(self):
        return self._get_default().__nonzero__()

    def __len__(self):
        return 1 if self._get_default() else 0

    def __str__(self):
        return str(self._get_default())

    def __iter__(self):
        return self._get_default().__iter__()


class TrueCallableProxy(CallableProxy):
    """ A version of CallableProxy that is always truthy """
    # pylint: disable=too-few-public-methods

    def __nonzero__(self):
        return True

    def __len__(self):
        return True

#: arrow format string for 'day' archives
DAY_FORMAT = 'YYYY-MM-DD'

#: arrow format string for 'month' archives
MONTH_FORMAT = 'YYYY-MM'

#: arrow format string for 'year' archives
YEAR_FORMAT = 'YYYY'


def parse_date(datestr):
    """ Parse a date expression into a tuple of:

        (start_date, span_type, span_format)

    Arguments:
        datestr -- A date specification, in the format of YYYY-MM-DD (dashes optional)
    """

    match = re.match(r'([0-9]{4})(-?([0-9]{1,2}))?(-?([0-9]{1,2}))?$', datestr)
    if not match:
        return (arrow.get(datestr,
                          tzinfo=config.timezone).replace(tzinfo=config.timezone),
                'day', 'YYYY-MM-DD')

    year, month, day = match.group(1, 3, 5)
    start = arrow.Arrow(year=int(year), month=int(
        month or 1), day=int(day or 1), tzinfo=config.timezone)

    if day:
        return start, 'day', DAY_FORMAT
    elif month:
        return start, 'month', MONTH_FORMAT
    elif year:
        return start, 'year', YEAR_FORMAT

    raise ValueError("Could not parse date: {}".format(datestr))


def find_file(path, search_path):
    """ Find a file by relative path. Arguments:

    path -- the image's filename
    search_path -- a list of directories to check in

    Returns: the resolved file path
    """

    if isinstance(search_path, str):
        search_path = [search_path]
    for relative in search_path:
        candidate = os.path.normpath(os.path.join(relative, path))
        if os.path.isfile(candidate):
            return candidate

    return None


def make_slug(title):
    """ convert a title into a URL-friendly slug """

    # https://github.com/fluffy-critter/Publ/issues/16
    # this should probably handle things other than English ASCII, and also
    # some punctuation should just be outright removed (quotes/apostrophes/etc)
    return re.sub(r"[^a-zA-Z0-9.]+", r" ", title).strip().replace(' ', '-')


def static_url(path, absolute=False):
    """ Shorthand for returning a URL for the requested static file.

    Arguments:

    path -- the path to the file (relative to the static files directory)
    absolute -- whether the link should be absolute or relative
    """
    return flask.url_for('static', filename=path, _external=absolute)



# rendering.py
""" Rendering functions """

from __future__ import absolute_import, with_statement

import os
import logging

import flask
from flask import request, redirect, render_template, url_for

from . import config
from . import path_alias, model
from .entry import Entry, expire_record
from .category import Category
from .template import Template
from .view import View
from . import caching
from .caching import cache

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# mapping from template extension to MIME type; probably could be better
EXTENSION_MAP = {
    '.html': 'text/html; charset=utf-8',
    '.xml': 'application/xml',
    '.json': 'application/json',
    '.css': 'text/css'
}


def mime_type(template):
    """ infer the content-type from the extension """
    _, ext = os.path.splitext(template.filename)
    return EXTENSION_MAP.get(ext, 'text/html; charset=utf-8')


@cache.memoize()
def map_template(category, template_list):
    """
    Given a file path and an acceptable list of templates, return the
    best-matching template's path relative to the configured template
    directory.

    Arguments:

    category -- The path to map
    template_list -- A template to look up (as a string), or a list of templates.
    """

    if isinstance(template_list, str):
        template_list = [template_list]

    for template in template_list:
        path = os.path.normpath(category)
        while path != None:
            for extension in ['', '.html', '.htm', '.xml', '.json']:
                candidate = os.path.join(path, template + extension)
                file_path = os.path.join(config.template_folder, candidate)
                if os.path.isfile(file_path):
                    return Template(template, candidate, file_path)
            parent = os.path.dirname(path)
            if parent != path:
                path = parent
            else:
                path = None


def get_redirect():
    """ Check to see if the current request is a redirection """
    return path_alias.get_redirect([request.full_path, request.path])


def render_error(category, error_message, error_codes, exception=None):
    """ Render an error page.

    Arguments:

    category -- The category of the request
    error_message -- The message to provide to the error template
    error_codes -- The applicable HTTP error code(s). Will usually be an
        integer or a list of integers; the HTTP error response will always
        be the first error code in the list, and the others are alternates
        for looking up the error template to use.
    exception -- Any exception that led to this error page
    """

    if isinstance(error_codes, int):
        error_codes = [error_codes]

    error_code = error_codes[0]
    template_list = [str(code) for code in error_codes]
    template_list.append('error')

    template = map_template(category, template_list)
    if template:
        return render_template(
            template.filename,
            error={'code': error_code, 'message': error_message},
            exception=exception,
            template=template), error_code

    # no template found, so fall back to default Flask handler
    return flask.abort(error_code)


def render_exception(error):
    """ Catch-all renderer for the top-level exception handler """
    _, _, category = str.partition(request.path, '/')
    return render_error(category, "Exception occurred", 500, exception={
        'type': type(error).__name__,
        'str': str(error),
        'args': error.args
    })


def render_path_alias(path):
    """ Render a known path-alias (used primarily for Dreamhost .php redirects) """

    redir = path_alias.get_redirect('/' + path)
    if not redir:
        return render_error('', 'Path redirection not found', 404)
    return redirect(redir)


@cache.cached(key_prefix=caching.make_category_key)
def render_category(category='', template='index'):
    """ Render a category page.

    Arguments:

    category -- The category to render
    template -- The template to render it with
    """
    # See if this is an aliased path
    redir = get_redirect()
    if redir:
        return redirect(redir)

    # Forbidden template types
    if template in ['entry', 'error']:
        return render_error(category, 'Unsupported template', 400)

    if category:
        # See if there's any entries for the view...
        if not model.Entry.get_or_none((model.Entry.category == category) |
                                       (model.Entry.category.startswith(category + '/'))):
            return render_error(category, 'Category not found', 404)

    tmpl = map_template(category, template)

    if not tmpl:
       # this might actually be a malformed category URL
        test_path = os.path.join(category, template)
        record = model.Entry.get_or_none(model.Entry.category == test_path)
        if record:
            return redirect(url_for('category', category=test_path))

        # nope, we just don't know what this is
        return render_error(category, 'Template not found', 400)

    view_spec = {'category': category}
    for key in ['date', 'last', 'first', 'before', 'after']:
        if key in request.args:
            view_spec[key] = request.args[key]

    view_obj = View(view_spec)
    return render_template(tmpl.filename,
                           category=Category(category),
                           view=view_obj,
                           template=tmpl), {'Content-Type': mime_type(tmpl)}


@cache.cached(key_prefix=caching.make_entry_key)
def render_entry(entry_id, slug_text='', category=''):  # pylint: disable=too-many-return-statements
    """ Render an entry page.

    Arguments:

    entry_id -- The numeric ID of the entry to render
    slug_text -- The expected URL slug text
    category -- The expected category
    """

    # check if it's a valid entry
    record = model.Entry.get_or_none(model.Entry.id == entry_id)
    if not record:
        # It's not a valid entry, so see if it's a redirection
        path_redirect = get_redirect()
        if path_redirect:
            return redirect(path_redirect)

        logger.info("Attempted to retrieve nonexistent entry %d", entry_id)
        return render_error(category, 'Entry not found', 404)

    # see if the file still exists
    if not os.path.isfile(record.file_path):
        expire_record(record)

        # See if there's a redirection
        path_redirect = get_redirect()
        if path_redirect:
            return redirect(path_redirect)

        return render_error(category, 'Entry not found', 404)

    # Show an access denied error if the entry has been set to draft mode
    if record.status == model.PublishStatus.DRAFT:
        return render_error(category, 'Entry not available', 403)

    # read the entry from disk
    entry_obj = Entry(record)

    # does the entry-id header mismatch? If so the old one is invalid
    if int(entry_obj.get('Entry-ID')) != record.id:
        expire_record(record)
        return render_error(category, 'Entry not found', 404)

    # check if the canonical URL matches
    if record.category != category or record.slug_text != slug_text:
        # This could still be a redirected path...
        path_redirect = get_redirect()
        if path_redirect:
            return redirect(path_redirect)

        # Redirect to the canonical URL
        return redirect(url_for('entry',
                                entry_id=entry_id,
                                category=record.category,
                                slug_text=record.slug_text))

    # if the entry canonically redirects, do that now
    entry_redirect = entry_obj.get('Redirect-To')
    if entry_redirect:
        return redirect(entry_redirect)

    tmpl = map_template(category, 'entry')
    if not tmpl:
        return render_error(category, 'Entry template not found', 400)

    return render_template(tmpl.filename,
                           entry=entry_obj,
                           category=Category(category),
                           template=tmpl), {'Content-Type': mime_type(tmpl)}



# category.py
""" The Category object passed to entry and category views """

from __future__ import absolute_import

import os

from flask import url_for

from . import model
from . import utils
from . import entry
from . import queries


class Category:
    """ Wrapper for category information """
    # pylint: disable=too-few-public-methods,too-many-instance-attributes

    def __init__(self, path):
        """ Initialize a category wrapper

        path -- the path to the category
        """

        self.path = path
        self.basename = os.path.basename(path)

        self._parent = os.path.dirname(path) if path else None

        subcat_query = model.Entry.select(model.Entry.category).distinct()
        if path:
            subcat_query = subcat_query.where(
                model.Entry.category.startswith(path + '/'))
        else:
            subcat_query = subcat_query.where(model.Entry.category != path)

        self._subcats_recursive = subcat_query.order_by(model.Entry.category)

        self.link = utils.CallableProxy(self._link)

        self.subcats = utils.CallableProxy(self._get_subcats)
        self.first = utils.CallableProxy(self._first)
        self.last = utils.CallableProxy(self._last)

    def _link(self, template='', absolute=False):
        return url_for('category',
                       category=self.path,
                       template=template,
                       _external=absolute)

    def __str__(self):
        return self.path

    def __getattr__(self, name):
        """ Lazily bind related objects """
        # pylint: disable=attribute-defined-outside-init
        if name == 'parent':
            self.parent = Category(self._parent) if (
                self._parent != None) else None
            return self.parent

        raise AttributeError("Unknown category attribute {}".format(name))

    def _get_subcats(self, recurse=False):
        """ Get the subcategories of this category

        recurse -- whether to include their subcategories as well
        """

        if recurse:
            # No need to filter
            return [Category(e.category) for e in self._subcats_recursive]

        # get all the subcategories, with only the first subdir added

        # number of path components to ingest
        parts = len(self.path.split('/')) + 1 if self.path else 1

        # get the subcategories
        subcats = [e.category for e in self._subcats_recursive]

        # convert them into separated pathlists with only 'parts' parts
        subcats = [c.split('/')[:parts] for c in subcats]

        # join them back into a path, and make unique
        subcats = {'/'.join(c) for c in subcats}

        # convert to a bunch of Category objects
        return [Category(c) for c in sorted(subcats)]

    def _entries(self, spec):
        """ Return a model query to get our entry records """
        return model.Entry.select().where(
            queries.build_query({**spec, 'category': self}))

    def _first(self, **spec):
        """ Get the earliest entry in this category, optionally including subcategories """
        first = self._entries(spec).order_by(
            model.Entry.entry_date, model.Entry.id)
        return entry.Entry(first[0]) if first else None

    def _last(self, **spec):
        """ Get the latest entry in this category, optionally including subcategories """
        last = self._entries(spec).order_by(
            -model.Entry.entry_date, -model.Entry.id)
        return entry.Entry(last[0]) if last else None
